/*
 * Copyright 2020 ETSI
 * 
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holder nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software without 
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package fr.emse.gitlab.saref.managers;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.apache.commons.io.FileUtils;
import org.apache.jena.query.Dataset;
import org.apache.jena.query.ReadWrite;
import org.apache.jena.rdf.model.ModelFactory;
import org.apache.jena.tdb.TDBFactory;
import org.slf4j.Logger;

import fr.emse.gitlab.saref.SAREF;
import fr.emse.gitlab.saref.SAREFErrorLogger;
import fr.emse.gitlab.saref.SAREFPipeline;
import fr.emse.gitlab.saref.SAREFPipelineException;

public class DatasetManager extends SAREFErrorLogger {

	private static final String NAME_DATASET = "tdb";
	private static final String CONFIG = SAREF.BASE + "config";

	private final File datasetDir;
	private final Dataset dataset;
	
	private static enum MESSAGE {
		bad_transaction, no_transaction
	}

	/**
	 * Prepare the Apache Jena TDB RDF dataset
	 */
	public DatasetManager(SAREFPipeline pipeline, Logger errorLogger) throws IOException {
		super(pipeline, errorLogger);
		datasetDir = new File(pipeline.targetDir, NAME_DATASET);
		FileUtils.forceMkdir(datasetDir);
		dataset = TDBFactory.createDataset(datasetDir.getAbsolutePath());
	}

	/**
	 * Empty the Apache Jena TDB RDF dataset
	 */
	public void emptyDataset() {
		dataset.begin(ReadWrite.WRITE);
		dataset.getDefaultModel().removeAll();
		List<String> toRemove = new ArrayList<>();
		Iterator<String> it = dataset.listNames();
		while (it.hasNext()) {
			toRemove.add(it.next());
		}
		for (String name : toRemove) {
			dataset.removeNamedModel(name);
		}
		dataset.addNamedModel(CONFIG, ModelFactory.createDefaultModel());
		dataset.commit();
	}

//	public void consumeDataset(Consumer<Dataset> consumer, ReadWrite readWrite) throws SAREFPipelineException {
//		boolean own = startTransaction(readWrite);
//		consumer.accept(dataset);
//		endTransaction(own);
//	}
//
//	public void consumeDataset(Consumer<Dataset> consumer) throws SAREFPipelineException {
//		consumeDataset(consumer, ReadWrite.READ);
//	}
//
//	public void consumeConfig(Consumer<Model> consumer, ReadWrite readWrite) throws SAREFPipelineException {
//		boolean own = startTransaction(readWrite);
//		final Model config = dataset.getNamedModel(CONFIG);
//		consumer.accept(config);
//		endTransaction(own);
//	}
//	
//	public Model getConfig(Dataset dataset) {
//		return dataset.getNamedModel(CONFIG);
//	}
//
//	public void consumeConfig(Consumer<Model> consumer) throws SAREFPipelineException {
//		consumeConfig(consumer, ReadWrite.READ);
//	}
//
//	public void setModel(SAREFVersion version, Model model) throws SAREFPipelineException {
//		boolean own = startTransaction(ReadWrite.WRITE);
//		dataset.addNamedModel(version.getVersionIRI(), model);
//		endTransaction(own);
//	}
//
//	public void setModel(SAREFExample example, Model model) throws SAREFPipelineException {
//		boolean own = startTransaction(ReadWrite.WRITE);
//		dataset.addNamedModel(example.getIRI(), model);
//		endTransaction(own);
//	}
//
//	public Model getModel(Dataset dataset, SAREFVersion version) {
//		return dataset.getNamedModel(version.getVersionIRI());
//	}
//
//	
//	public void consumeModel(SAREFVersion version, Consumer<Model> consumer, ReadWrite readWrite) throws SAREFPipelineException {
//		boolean own = startTransaction(readWrite);
//		Model model = dataset.getNamedModel(version.getVersionIRI());
//		consumer.accept(model);
//		endTransaction(own);		
//	}
//	
//	public void consumeModel(SAREFVersion version, Consumer<Model> consumer) throws SAREFPipelineException {
//		consumeModel(version, consumer, ReadWrite.READ);
//	}
//
//	public Model getModel(Dataset dataset, SAREFExample example) {
//		return dataset.getNamedModel(example.getIRI());
//	}
//	
//	public void consumeModel(SAREFExample example, Consumer<Model> consumer, ReadWrite readWrite) throws SAREFPipelineException {
//		boolean own = startTransaction(readWrite);
//		Model model = dataset.getNamedModel(example.getIRI());
//		consumer.accept(model);
//		endTransaction(own);	
//	}
//	
//	public void consumeModel(SAREFExample example, Consumer<Model> consumer) throws SAREFPipelineException {
//		consumeModel(example, consumer, ReadWrite.READ);
//	}
//
//	
//	public void consumeModel(SAREFTerm term, Consumer<Model> consumer, ReadWrite readWrite) throws SAREFPipelineException {
//		boolean own = startTransaction(readWrite);
//		Model model = dataset.getNamedModel(term.getIRI());
//		consumer.accept(model);
//		endTransaction(own);	
//	}
//	
//	public void consumeModel(SAREFTerm term, Consumer<Model> consumer) throws SAREFPipelineException {
//		consumeModel(term, consumer, ReadWrite.READ);
//	}
//	
//	public void writeConfig() {
//		File configFile = new File(pipeline.targetDir, "config.ttl");
//		try (FileOutputStream fos = new FileOutputStream(configFile)) {
//			boolean own = startTransaction(ReadWrite.READ);
//			consumeConfig(config -> config.write(fos, Lang.TURTLE.getLabel()));
//			endTransaction(own);
//		} catch (SAREFPipelineException | IOException ex) {
//			errorLogger.error("Unexpected exception while writing configuration file", ex);
//		}
//	}

	private boolean startTransaction(ReadWrite readWrite) throws SAREFPipelineException {
		if (dataset.transactionMode() == null) {
			dataset.begin(readWrite);
			return true;
		} else if(dataset.transactionMode() != readWrite) {
			throw new SAREFPipelineException(getMessage(MESSAGE.bad_transaction, readWrite));
		}
		return false;
	}

	private void endTransaction(boolean own) throws SAREFPipelineException {
		if (dataset.transactionMode() == null) {
			throw new SAREFPipelineException(getMessage(MESSAGE.no_transaction));
		} else if(dataset.transactionMode() == ReadWrite.WRITE) {
			dataset.commit();
		} 
		if(own) {
			dataset.end();
		}
	}

}
